"""Exceptions"""


class InvalidUsage(Exception):
    """Badly-formed payload."""

    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = message
        self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        return_value = dict(self.payload or ())
        return_value["status code"] = self.status_code
        return_value["message"] = self.message
        return return_value



"""
Main package.

Define the end points of the REST API; delegate responses to strategies
"""
import json
import os

from flask import Flask, jsonify, request

from . import bamboo
from . import context
from . import helpers
from . import strategies
from . import exceptions
from . import jsonparser

USER_HOME = os.environ["HOME"]

try:
    NETRC_FILE = os.environ["NETRC_FILE"]
except KeyError:
    print("ERROR: Location of netrc file for Bamboo credentials not specified")
    NETRC_FILE = None

try:
    BAMB_HOSTNAME = os.environ["BAMB_HOSTNAME"]
except KeyError:
    print("ERROR: Bamboo hostname not specified")
    BAMB_HOSTNAME = None

try:
    BITB_HOSTNAME = os.environ["BITB_HOSTNAME"]
except KeyError:
    print("ERROR: Bitbucket hostname not specified")
    BITB_HOSTNAME = None

try:
    BITB_PROJECT = os.environ["BITB_PROJECT"]
except KeyError:
    print("ERROR: Bitbucket project name not specified")
    BITB_PROJECT = None

try:
    BITB_REPO = os.environ["BITB_REPO"]
except KeyError:
    print("ERROR: Bitbucket repository name not specified")
    BITB_REPO = None

conns = {}
try:
    conns["bamb"] = bamboo.BambooConnection(NETRC_FILE, BAMB_HOSTNAME)
except FileNotFoundError:
    print("ERROR: netrc file not found")
    conns["bamb"] = None

try:
    conns["bitb"] = bamboo.BambooConnection(NETRC_FILE, BITB_HOSTNAME)
except FileNotFoundError:
    print("ERROR: netrc file not found")
    conns["bitb"] = None

## TODO: temporary checks until in routine use --------------------------------
## Check contact with Bamboo. (Just make a simple GET API call that lists all projects.)
# response = conns["bamb"].get("project.json")
# assert response.status_code == 200
# helpers.log(json.dumps(response.json(), sort_keys=True, indent=4))
## Check contact with Bitbucket. (Just make a simple GET API call that lists all projects.)
# response = conns["bitb"].get("projects", apiversion="1.0")
# assert response.status_code == 200
# helpers.log(json.dumps(response.json(), sort_keys=True, indent=4))
# ----------------------------------------------------------------------------


app = Flask(__name__)


@app.errorhandler(exceptions.InvalidUsage)
def handle_invalid_usage(error):
    """Error handler for InvalidUsage exception."""
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    return response


@app.route("/flafl/api/v1.0/events", methods=["POST"])
def post_event():
    debug_info = {}
    debug_info["payloadReceived"] = request.json

    # Verbose
    # helpers.log(json.dumps(request.json, sort_keys=True, indent=4))

    # For 'Test Connection' function in Bamboo webhook settings
    if jsonparser.is_connection_test(request.json):
        message = "Successful connection."
        return jsonify({"message": message, "debug_info": debug_info})

    event_key, event_trigger = jsonparser.get_event_key(request.json, debug_info)

    debug_info["eventKey"] = event_key
    debug_info["eventTrigger"] = event_trigger

    # Select the appropriate strategy
    if event_key.startswith("pr:"):
        if event_trigger == "opened":
            concrete_strategy = strategies.PrCreate()

        if event_trigger == "deleted":
            concrete_strategy = strategies.PrDestroy()

        if event_trigger == "merged":
            concrete_strategy = strategies.PrDestroy()

        if event_trigger == "declined":
            concrete_strategy = strategies.PrDestroy()

        if event_trigger == "modified":
            concrete_strategy = strategies.PrModify()

        if event_trigger == "comment":
            concrete_strategy = strategies.PrComment()

    if event_key.startswith("repo:"):
        if event_trigger == "refs_changed":
            concrete_strategy = strategies.RepoPush()

        else:
            concrete_strategy = strategies.RepoAll()

    # Execute the strategy
    try:
        ct = context.Context(concrete_strategy)
    except UnboundLocalError:
        message = "Trigger-handling for eventKey not coded yet"
        payload = {"eventKey": event_key}
        success_payload = {"message": message, "payload": payload}
    else:
        return_value = ct.execute_strategy(request.json, debug_info, conns)
        success_payload = return_value.get_json()

    try:
        helpers.log(
            "Success payload:\n" + json.dumps(success_payload, sort_keys=True, indent=4)
        )
    except TypeError:
        pass

    return jsonify(success_payload)



"""Context module."""


class Context:
    """Define the interface of interest to clients."""

    def __init__(self, strategy):
        """Maintain a reference to a Strategy object."""
        self._strategy = strategy

    def execute_strategy(self, json_data, debug_info, conns):
        """Execute wrapper for the Strategy object's execute method."""
        return_value = self._strategy.execute(json_data, debug_info, conns)
        return return_value



"""Helper functions for parsing JSON payloads from POST requests to the API."""

from . import exceptions

INVALID_USAGE = exceptions.InvalidUsage


def is_connection_test(json_data):
    """Check if event is just a connection test"""
    return "test" in json_data


def get_event_key(json_data, debug_info):
    """Get eventKey.

    The eventKey is the primary key that defines the event that triggered a
    webhook.
    """
    if "eventKey" not in json_data:
        message = "POST method to this endpoint must provide an eventKey"
        raise INVALID_USAGE(message, status_code=410, payload=debug_info)
    event_key = json_data["eventKey"]
    event_key_field = event_key.split(":")
    if len(event_key_field) < 2:
        message = "eventKey must contain two or more colon-separated items"
        raise INVALID_USAGE(message, status_code=410, payload=debug_info)
    event_trigger = event_key_field[1]
    return event_key, event_trigger


def verify_pull_request_id(json_data, debug_info):
    """Look for pullRequest id key."""
    try:
        pull_request_id = str(json_data["pullRequest"]["id"])
    except TypeError:
        message = "Pull request event does not contain pullRequest id key"
        raise INVALID_USAGE(message, status_code=420, payload=debug_info)
    except KeyError:
        message = "Pull request event does not contain pullRequest key"
        raise INVALID_USAGE(message, status_code=420, payload=debug_info)
    if not pull_request_id.isdigit():
        message = "Pull Request ID is not an integer"
        raise INVALID_USAGE(message, status_code=420, payload=debug_info)
    return pull_request_id


def get_from_ref_latest_commit(json_data, debug_info):
    """Get latestCommit hash for the fromRef branch."""
    try:
        from_ref_latest_commit = json_data["pullRequest"]["fromRef"]["latestCommit"]
    except KeyError:
        message = "Pull request event does not contain fromRef key"
        raise INVALID_USAGE(message, status_code=420, payload=debug_info)
    except TypeError:
        message = "Pull request fromRef does not contain latestCommit key"
        raise INVALID_USAGE(message, status_code=420, payload=debug_info)
    if not isinstance(from_ref_latest_commit, str) or len(from_ref_latest_commit) != 40:
        message = "Pull request fromRef latestCommit not a 40-character string"
        debug_info["type_of_from_ref_latest_commit"] = str(type(from_ref_latest_commit))
        debug_info["len_of_from_ref_latest_commit"] = len(from_ref_latest_commit)
        raise INVALID_USAGE(message, status_code=420, payload=debug_info)
    return from_ref_latest_commit



import abc
import yaml
from flask import jsonify

# import subprocess

from . import exceptions
from . import helpers
from . import jsonparser

INVALID_USAGE = exceptions.InvalidUsage


class Strategy:
    """
    Declare an interface common to all supported algorithms. Context
    uses this interface to call the algorithm defined by a
    concrete strategy.
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def execute(self, json_data, debug_info, conns):
        pass


class PrCreate(Strategy):
    """Class for pull-request creation response."""

    def execute(self, json_data, debug_info, conns):
        jsonparser.verify_pull_request_id(json_data, debug_info)

        # Get the required info from the PR webhook payload
        pull_request_id = json_data["pullRequest"]["id"]
        pull_request_title = json_data["pullRequest"]["title"]
        to_ref_slug = json_data["pullRequest"]["toRef"]["repository"]["slug"]
        to_ref_proj = json_data["pullRequest"]["toRef"]["repository"]["project"]["key"]

        # Run basic checks on the pull request
        helpers.check_for_jira_ticket(
            conns, to_ref_proj, to_ref_slug, pull_request_id, pull_request_title
        )

        # Add the new plan (for this PR) to the Bamboo specs
        specs = {
            "version": 2,
            "plan": {
                "project-key": "bamboo-specs-repo",
                "key": to_ref_slug + str(pull_request_id),
                "name": pull_request_title,
            },
            "stages": [{"Stage 1": {"jobs": ["Job 1"]}}],
            "Job 1": {"tasks": [{"script": ["echo 'Test YAML spec'"]}]},
        }
        helpers.log(yaml.dump(specs, default_flow_style=False))

        # Trigger the scan of the Bamboo specs repo
        try:
            response = conns["bamb"].get("project/bamboo-specs-repo/repository")
            api_message = "API call to Bamboo not made."
        except AttributeError:
            print("ERROR: cannot scan Bamboo specs repository")
            api_message = "API call to Bamboo not made."
        else:
            helpers.log(
                "Repos with access to create plans in project bamboo-specs-repo:\n"
                + response.text
            )
            response = conns["bamb"].post(
                "repository/scan?repositoryName=test_bamboo_specs"
            )
            helpers.log("Response from Bamboo scan:\n" + response.text)
            response = conns["bamb"].post("repository/scan?repositoryId=2673")
            helpers.log("Response from Bamboo scan:\n" + response.text)
            api_message = "Sent API call to Bamboo and got return code " + str(
                response.status_code
            )

        message = (
            "Created PR with ID "
            + str(pull_request_id)
            + " from "
            + json_data["pullRequest"]["fromRef"]["repository"]["project"]["key"]
            + "/"
            + json_data["pullRequest"]["fromRef"]["repository"]["slug"]
            + "/"
            + json_data["pullRequest"]["fromRef"]["displayId"]
            + " to "
            + json_data["pullRequest"]["toRef"]["repository"]["project"]["key"]
            + "/"
            + json_data["pullRequest"]["toRef"]["repository"]["slug"]
            + "/"
            + json_data["pullRequest"]["toRef"]["displayId"]
            + ". "
            + api_message
        )

        return jsonify({"message": message, "status": "success"})


class PrDestroy(Strategy):
    """Class for pull-request closure response."""

    def execute(self, json_data, debug_info, conns):
        jsonparser.verify_pull_request_id(json_data, debug_info)

        message = (
            "Destroyed PR with ID "
            + str(json_data["pullRequest"]["id"])
            + " in repository "
            + json_data["pullRequest"]["fromRef"]["repository"]["project"]["key"]
            + "/"
            + json_data["pullRequest"]["toRef"]["repository"]["slug"]
        )
        return jsonify({"message": message, "status": "success"})


class PrModify(Strategy):
    """Class for pull-request modification response."""

    def execute(self, json_data, debug_info, conns):
        jsonparser.verify_pull_request_id(json_data, debug_info)

        pull_request_id = json_data["pullRequest"]["id"]
        pull_request_title = json_data["pullRequest"]["title"]
        to_ref_slug = json_data["pullRequest"]["toRef"]["repository"]["slug"]
        to_ref_proj = json_data["pullRequest"]["toRef"]["repository"]["project"]["key"]

        helpers.check_for_jira_ticket(
            conns, to_ref_proj, to_ref_slug, pull_request_id, pull_request_title
        )

        # Get the fromRef latestCommit
        from_ref_latest_commit = jsonparser.get_from_ref_latest_commit(
            json_data, debug_info
        )

        event_key, event_trigger = jsonparser.get_event_key(json_data, debug_info)
        message = (
            "status:"
            + "success"
            + "from_ref_latest_commit:"
            + from_ref_latest_commit
            + "eventKey:"
            + event_key
            + "pull_request_id:"
            + ""
            + "eventTrigger:"
            + event_trigger
        )
        return jsonify(
            {"message": message, "status": "success", "payload": from_ref_latest_commit}
        )


class PrComment(Strategy):
    """Class for pull-request comment responses."""

    def execute(self, json_data, debug_info, conns):
        """Strategy for pull-request comment events."""
        jsonparser.verify_pull_request_id(json_data, debug_info)

        # Check for comment field in JSON
        if "comment" not in json_data:
            message = "Payload for comment event did not contain comment key"
            debug_info["receivedPayload"] = json_data
            raise INVALID_USAGE(message, status_code=410, payload=debug_info)
        if "author" not in json_data["comment"]:
            message = "Payload for comment event did not contain author key"
            debug_info["receivedPayload"] = json_data
            raise INVALID_USAGE(message, status_code=410, payload=debug_info)
        if "name" not in json_data["comment"]["author"]:
            message = "Payload for comment event did not contain name key"
            debug_info["receivedPayload"] = json_data
            raise INVALID_USAGE(message, status_code=410, payload=debug_info)

        message = "Comment by " + json_data["comment"]["author"]["name"]

        # Do something...

        return jsonify({"message": message, "status": "success"})


class RepoPush(Strategy):
    """Class for repository-based webhook responses."""

    def execute(self, json_data, debug_info, conns):
        """Strategy for repository-based webhooks."""
        event_key, _ = jsonparser.get_event_key(json_data, debug_info)
        message = "Repository received push"
        success_payload = {
            "eventKey": event_key,
            "project": json_data["repository"]["project"]["key"],
        }
        return jsonify({"message": message, "payload": success_payload})


class RepoAll(Strategy):
    """Class for repository-based webhook responses."""

    def execute(self, json_data, debug_info, conns):
        """Strategy for repository-based webhooks."""
        event_key, _ = jsonparser.get_event_key(json_data, debug_info)
        message = "Trigger-handling for eventKey not coded yet"
        success_payload = {"eventKey": event_key}
        return jsonify({"message": message, "payload": success_payload})



import datetime  # for log only
import re


def log(log_entry):
    with open(__package__ + ".log", "a") as file_pointer:
        file_pointer.write(str(datetime.datetime.now()) + "\n")
        file_pointer.write(log_entry)
        file_pointer.write("\n")


def check_for_jira_ticket(
    conns, to_ref_proj, to_ref_slug, pull_request_id, pull_request_title
):
    JIRA_TICKET_REGEX = re.compile(to_ref_proj + "-[0-9]+")
    # PR ID 1 is used for testing
    if pull_request_id != "1" and not JIRA_TICKET_REGEX.match(pull_request_title):
        comment = (
            "Please add a Jira ticket ID to the pull request title. E.g. 'JIRA-123: "
            + pull_request_title
            + "'"
        )
        try:
            conns["bitb"].post(
                "projects/"
                + to_ref_proj
                + "/repos/"
                + to_ref_slug
                + "/pull-requests/"
                + str(pull_request_id)
                + "/comments",
                json={"text": comment},
            )
        except AttributeError:
            print("ERROR: no Bitbucket connection")



import netrc
import requests


class BambooConnection:
    """Create connection to Atlassian tool, for API calls"""

    def __init__(self, netrc_file, hostname):
        """Create the connection, using the provided netrc file"""
        netrc_object = netrc.netrc(netrc_file)
        username = netrc_object.authenticators(hostname)[0]
        password = netrc_object.authenticators(hostname)[2]
        self.auth = (username, password)
        self.url = "https://" + hostname + "/rest/api"

    def get(self, resource, apiversion="latest"):
        """Issue a GET API call"""
        response = requests.get(
            url=self.url + "/" + apiversion + "/" + resource, auth=self.auth
        )
        return response

    def post(self, resource, apiversion="latest", json=None):
        """Issue a POST API call with a JSON payload"""
        response = requests.post(
            url=self.url + "/" + apiversion + "/" + resource, auth=self.auth, json=json
        )
        return response
